package com.tomwheeler.stocktrader.api;

import java.util.Collection;

import com.tomwheeler.stocktrader.api.events.NewsStoryListener;
import com.tomwheeler.stocktrader.api.events.StockPriceListener;

/**
 * <p>
 * This interface represents services offered by an entity that provides news
 * and statistics about the stock market.   
 * </p>
 * 
 * @author Tom Wheeler
 */
public interface StockDataProvider {

    /**
     * <p>
     * Returns the price for the specified stock (based on its ticker symbol) on
     * the specified day.
     * </p>
     * 
     * @param symbol
     *            the stock symbol for which you want the price
     * @param dayNumber
     *            the day for which you want the price
     * @return the price of stock on the specified day
     * @throws IllegalArgumentException
     *            if the <code>symbol</code> is invalid (not listed on the
     *            exchange), or if the <code>dayNumber</code> is out of range
     *            (less than one or past the current day in the game).
     */
    public int getPrice(String symbol, int dayNumber) throws IllegalArgumentException;

    /**
     * <p>
     * Returns a collection (of Strings) representing market news for the
     * specified day.
     * </p>
     * 
     * @param dayNumber
     *           the day for which you want the news
     * @return a collection of Strings, each of which contains a news item
     */
    public Collection getNewsItems(int dayNumber);
    
    /** 
     * <p>
     * This method is called by the Game to indicate that a new trading day has started
     * and that all daily initialization tasks, such as generating market news and 
     * stock prices, should be performed.
     * </p>
     */
    public void performDailyInitialization();
    
    /** 
     * <p>
     * Registers a listener with this provider so that it can be notified of market
     * news stories being published.
     * </p> 
     * 
     * @param listener the news story listener being registered
     */
    public void addNewsStoryListener(NewsStoryListener listener);
    
    /** 
     * <p>
     * Registers a listener with this provider so that it can be notified of updates
     * to stock prices.
     * </p> 
     * 
     * @param listener the stock price listener being registered
     */
    public void addStockPriceListener(StockPriceListener listener);
}
