package com.tomwheeler.stocktrader.engine;


import java.text.NumberFormat;
import java.util.*;

import com.tomwheeler.stocktrader.api.*;

/**
 * <p>
 * This is a concrete implementation of the StockExchange interface, used for
 * the stocktrader game.
 * </p>
 *  
 * @author Tom Wheeler
 */
public class StockExchangeImpl implements StockExchange {

    private static Collection stockList = createListedStocks();
    
    private Game game;
    private static NumberFormat currencyFormatter;

    public StockExchangeImpl(Game game)  {
        currencyFormatter = NumberFormat.getCurrencyInstance();
        this.game = game;
    }

    /* (non-Javadoc)
     * @see com.tomwheeler.stocktrader.engine.api.StockExchange#getAvailableStocks()
     */
    public Collection getListedStocks() {
        return Collections.unmodifiableCollection(stockList);
    }
    
    /* (non-Javadoc)
     * @see com.tomwheeler.stocktrader.engine.api.StockExchange#getListedStocks(com.tomwheeler.stocktrader.engine.api.Industry)
     */
    public Collection getListedStocks(Industry industry) {
        Collection industryStocks = new ArrayList();
        for (Iterator iter = stockList.iterator(); iter.hasNext();) {
            Stock stock = (Stock) iter.next();
            if (stock.getIndustry().equals(industry)) {
                industryStocks.add(stock);
            }
        }
        return Collections.unmodifiableCollection(industryStocks);
    }
    
    /* (non-Javadoc)
     * @see com.tomwheeler.stocktrader.engine.api.StockExchange#getStock(java.lang.String)
     */
    public Stock getStock(String symbol) throws IllegalArgumentException {
        for (Iterator iter = stockList.iterator(); iter.hasNext();) {
            Stock stock = (Stock) iter.next();
            if (stock.getSymbol().equals(symbol)) {
                return stock;
            }
        }

        String msg = "No stock with symbol '" + symbol + "' is listed on this exchange";
        throw new IllegalArgumentException(msg);
    }

    /* (non-Javadoc)
     * @see com.tomwheeler.stocktrader.engine.api.StockExchange#buy(com.tomwheeler.stocktrader.engine.api.StockAccount, java.lang.String, int)
     */
    public String buy(StockAccount acct, String symbol, int qty) throws IllegalArgumentException {
        Stock stock = getStock(symbol);  // throws IAE if not listed
        if (acct == null) {
            throw new IllegalArgumentException("Account cannot be null when buying stocks");            
        }
        if (qty <= 0) {
            throw new IllegalArgumentException("Quantity must be at least 1 share");            
        }

        int currentDay = game.getCurrentDay();
        int currentPrice = game.getDataProvider().getPrice(symbol, currentDay);
        int totalPrice = (qty * currentPrice);

        int cashBalance = acct.getCashBalance();
        if (cashBalance < totalPrice) {
            String msg = "Your account has insufficient cash for this purchase. " 
                       + "You have " + currencyFormatter.format(cashBalance) + " but "
                       + "need " + currencyFormatter.format(totalPrice) + ".";
            throw new IllegalArgumentException(msg);
        }
        
        StringBuffer confMsgBuff = new StringBuffer();
        confMsgBuff.append("You purchased " + qty);
        if (qty == 1) {
            confMsgBuff.append(" share of '");    
        } else {
            confMsgBuff.append(" shares of '");            
        }        
        confMsgBuff.append(symbol);
        confMsgBuff.append("' @ " + currencyFormatter.format(currentPrice));
        confMsgBuff.append(" each on day #");
        confMsgBuff.append(currentDay);
        confMsgBuff.append(" for a total price of ");
        confMsgBuff.append(currencyFormatter.format(totalPrice));
        
        return confMsgBuff.toString();
    }

    /* (non-Javadoc)
     * @see com.tomwheeler.stocktrader.engine.api.StockExchange#sell(com.tomwheeler.stocktrader.engine.api.StockAccount, java.lang.String, int)
     */
    public String sell(StockAccount acct, String symbol, int qty) throws IllegalArgumentException {        
        Stock stock = getStock(symbol);  // throws IAE if not listed
        if (acct == null) {
            throw new IllegalArgumentException("Account cannot be null when selling stocks");            
        }
        if (qty <= 0) {
            throw new IllegalArgumentException("Quantity must be >= 1 share");
        }
        
        int sharesHeld = acct.getQuantity(stock);
        if (sharesHeld < qty) {
            String msg = "You cannot sell " + qty + " shares of '" + stock.getSymbol()
                         + "' when your account only has " + sharesHeld + " shares.";                         
            throw new IllegalArgumentException(msg);
        }

        int currentDay = game.getCurrentDay();
        int currentPrice = game.getDataProvider().getPrice(symbol, currentDay);
        
        StringBuffer confMsgBuff = new StringBuffer();
        confMsgBuff.append("You sold " + qty);
        if (qty == 1) {
            confMsgBuff.append(" share of '");    
        } else {
            confMsgBuff.append(" shares of '");            
        }     
        confMsgBuff.append(symbol);
        confMsgBuff.append("' @ ");
        confMsgBuff.append(currencyFormatter.format((long)currentPrice));
        confMsgBuff.append(" on day ");
        confMsgBuff.append(currentDay);
        confMsgBuff.append(" for a total price of ");
        confMsgBuff.append(currencyFormatter.format(qty * currentPrice));

        return confMsgBuff.toString();
    }
    
    /** creates the stocks and lists them on the exchange */
    private static Collection createListedStocks() {
        Collection stocks = new HashSet();

        stocks.add(new StockImpl("FFC", "Fuel For Cheap", Industry.ENERGY));
        stocks.add(new StockImpl("ACM", "ACME Health Company", Industry.HEALTHCARE));
        stocks.add(new StockImpl("SSN", "Straight Scoop News", Industry.MEDIA));
        stocks.add(new StockImpl("QSL", "QuickShip Lines", Industry.TRANSPORTATION));
        stocks.add(new StockImpl("DCI", "Dewey-Cheatham Investments", Industry.BANKING));

        stocks.add(new StockImpl("THH", "Tiper-Hension Health Systems", Industry.HEALTHCARE));
        stocks.add(new StockImpl("FEF", "Factory Equipment Fabricators", Industry.MANUFACTURING));
        stocks.add(new StockImpl("WND", "Waters Indy Fuel", Industry.ENERGY));
        stocks.add(new StockImpl("HMI", "Andato Domani Surety", Industry.INSURANCE));
        stocks.add(new StockImpl("AVT", "Alambre Viejo Telecom", Industry.TELECOMMUNICATIONS));
        
        stocks.add(new StockImpl("LGH", "Lande-Grabb Holdings", Industry.REAL_ESTATE));
        stocks.add(new StockImpl("CMB", "Cash Money Bancorp", Industry.BANKING));        
        stocks.add(new StockImpl("SPT", "Snailpace Trucking", Industry.TRANSPORTATION));
        stocks.add(new StockImpl("GEC", "Grande Entertainment Conglomerate", Industry.MEDIA));
        stocks.add(new StockImpl("KNE", "Knicknack Emporium", Industry.RETAIL));
        
        stocks.add(new StockImpl("DME", "Digga-Mupp Excavation", Industry.MINING));
        stocks.add(new StockImpl("TBS", "Two-Bit Systems", Industry.ELECTRONICS));
        stocks.add(new StockImpl("BOI", "Bigg-Olde Industries", Industry.MANUFACTURING));        
        stocks.add(new StockImpl("NGM", "Notsa-Goode Hospital Management", Industry.HEALTHCARE));
        stocks.add(new StockImpl("BCC", "Bookcooker Communications", Industry.TELECOMMUNICATIONS));
        
        stocks.add(new StockImpl("BGF", "Big Green Financial", Industry.BANKING));        
        stocks.add(new StockImpl("GIT", "Get-it-There Transportation", Industry.TRANSPORTATION));
        stocks.add(new StockImpl("WEM", "Wooly-Eye Media.", Industry.MEDIA));
        stocks.add(new StockImpl("SFA", "Sprawl For All", Industry.REAL_ESTATE));
        stocks.add(new StockImpl("RMG", "Rusty Machine Group", Industry.MANUFACTURING));
        
        stocks.add(new StockImpl("BBS", "Big Box Stores", Industry.RETAIL));
        stocks.add(new StockImpl("GED", "Gate Estate Development", Industry.REAL_ESTATE));
        stocks.add(new StockImpl("OHC", "Ober-Heipt Communications", Industry.TELECOMMUNICATIONS));
        stocks.add(new StockImpl("NPE", "Nullpointer Computers", Industry.ELECTRONICS));
        stocks.add(new StockImpl("CPI", "Cash-Pitt Insurance", Industry.INSURANCE));
        
        stocks.add(new StockImpl("SAS", "Snazzy Apparel Shops", Industry.RETAIL));
        stocks.add(new StockImpl("NRG", "NRG Fuel Distribution", Industry.ENERGY));        
        stocks.add(new StockImpl("FSC", "FS-Check Storage Arrays", Industry.ELECTRONICS));
        stocks.add(new StockImpl("SCI", "Smallclaim Insurance", Industry.INSURANCE));
        stocks.add(new StockImpl("DHM", "Deep Hole Mining", Industry.MINING));

        return stocks;
    }
}
