package com.tomwheeler.stocktrader.ui.exchange;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.NumberFormat;
import java.util.*;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableColumnModel;

import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import org.openide.awt.StatusDisplayer;
import org.openide.util.NbBundle;

import com.tomwheeler.stocktrader.api.*;
import com.tomwheeler.stocktrader.api.events.*;
import com.tomwheeler.stocktrader.ui.StockListTableModel;
import com.tomwheeler.stocktrader.ui.renderer.IndustryCellRenderer;
import com.tomwheeler.stocktrader.ui.renderer.PercentChangeCellRenderer;

/**
 * <p>
 * This class provides a component that displays all stocks listed
 * on the exchange.  It provides the main part of the UI rendered
 * by the ExchangeListingTopComponent.  Aside from how user notification 
 * and resource bundles are handled, there's nothing specific to NetBeans
 * in this class.
 * </p>
 * 
 * @author Tom Wheeler
 */
public class ExchangeListingComponent extends JPanel {

    private static NumberFormat priceFormatter = NumberFormat.getCurrencyInstance();
    
    private static final int SYMBOL_COL_IDX = 0;
    private static final int NAME_COL_IDX = 1;
    private static final int INDUSTRY_COL_IDX = 2;
    private static final int CURRENT_PRICE_COL_IDX = 3;
    private static final int PRICE_CHANGE_PCT_COL_IDX = 4;
    
    private JTable stockTable;
    private StockListTableModel stockTableModel;
    private JLabel cashBalanceLabel;
    private JButton buyButton;    
    private StockPriceListener priceListener;

    private Game game;

    public ExchangeListingComponent(final Game game) {
        super(new BorderLayout());
        
        this.game = game;
        game.addGameListener(new CurrentGameListener());
        
        stockTableModel = new StockListTableModel();
        stockTableModel.addColumn(NbBundle.getMessage(ExchangeListingComponent.class,
                "LBL_SymbolCol"));
        stockTableModel.addColumn(NbBundle.getMessage(ExchangeListingComponent.class,
                "LBL_NameCol"));
        stockTableModel.addColumn(NbBundle.getMessage(ExchangeListingComponent.class,
                "LBL_IndustryCol"));
        stockTableModel.addColumn(NbBundle.getMessage(ExchangeListingComponent.class,
                "LBL_PriceCol"));
        stockTableModel.addColumn(NbBundle.getMessage(ExchangeListingComponent.class,
                "LBL_ChangePctCol"));

        stockTable = new JTable(stockTableModel);
        ListSelectionModel singleSelectionModel = new DefaultListSelectionModel();
        singleSelectionModel.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        stockTable.setSelectionModel(singleSelectionModel);
        stockTable.getTableHeader().setReorderingAllowed(false); 
        
        TableColumnModel colModel = stockTable.getColumnModel();
        colModel.getColumn(INDUSTRY_COL_IDX).setCellRenderer(new IndustryCellRenderer());
        colModel.getColumn(PRICE_CHANGE_PCT_COL_IDX).setCellRenderer(new PercentChangeCellRenderer());
        
        JScrollPane scroller = new JScrollPane(stockTable,
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        add(scroller, BorderLayout.CENTER);
        
        buyButton = new JButton(NbBundle.getMessage(ExchangeListingComponent.class, "LBL_BuyButton"));
        buyButton.setEnabled(false);  // enabled upon selection
        add(buyButton, BorderLayout.SOUTH);

        stockTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                if (game.isActive()) {
                    boolean isEnabled = stockTable.getSelectedRow() != -1;
                    buyButton.setEnabled(isEnabled);                    
                }
            }            
        });

        buyButton.addActionListener(new BuyStockActionListener());

        populateListingTable();
        priceListener = new PriceListener();
    }    
    
    /** attempts to buy the specified stock and handles confirmation/error message display */
    private void attemptPurchase(String symbol, int qty) {
        try {
            String confMsg = game.getAccount().buy(symbol, qty);
            StatusDisplayer.getDefault().setStatusText(confMsg);
        } catch (IllegalArgumentException iae) {
            String msg = iae.getLocalizedMessage();
            NotifyDescriptor descriptor = new NotifyDescriptor.Message(msg,
                    NotifyDescriptor.ERROR_MESSAGE);
            DialogDisplayer.getDefault().notify(descriptor);
        }
    }
    
    /** populates the table with data from all stocks listed on the exchange */
    private void populateListingTable() {
        if (! game.isActive()) {
            return;
        }
        
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    populateListingTable();
                }
            });
        }

        StockDataProvider provider = game.getDataProvider();        
        int dayNum = game.getCurrentDay();

        Collection allStocks = game.getExchange().getListedStocks();
        for (Iterator iter = allStocks.iterator(); iter.hasNext();) {
            Stock stock = (Stock) iter.next();
            int todayPrice = provider.getPrice(stock.getSymbol(), dayNum);
            Float priceChangePct = null;

            if (dayNum > 1) {
                // no change on the first day since there's nothing to compare to
                int lastPrice = provider.getPrice(stock.getSymbol(), dayNum - 1);
                priceChangePct = new Float((float)(todayPrice - lastPrice) / lastPrice);
            }

            Vector rowData = new Vector();
            rowData.insertElementAt(stock.getSymbol(), SYMBOL_COL_IDX);
            rowData.insertElementAt(stock.getName(), NAME_COL_IDX);
            rowData.insertElementAt(stock.getIndustry(), INDUSTRY_COL_IDX);
            rowData.insertElementAt(new Integer(todayPrice), CURRENT_PRICE_COL_IDX);
            rowData.insertElementAt(priceChangePct, PRICE_CHANGE_PCT_COL_IDX);
            stockTableModel.addRow(rowData);
        }
    }
    
    /** updates the listing for the specified stock */
    private void updateStockListing(final String stockSymbol) {
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    updateStockListing(stockSymbol);
                }
            });
        }
        
        StockDataProvider provider = game.getDataProvider();
        int dayNum = game.getCurrentDay();
        int todayPrice = provider.getPrice(stockSymbol, dayNum);
        Float priceChangePct = null;
        if (dayNum > 1) {
            int lastPrice = provider.getPrice(stockSymbol, dayNum - 1);
            priceChangePct = new Float((float)(todayPrice - lastPrice) / lastPrice);
        }
        
        int numRows = stockTableModel.getRowCount();
        for (int i = 0; i < numRows; i++) {
            String rowSymbol = (String)stockTableModel.getValueAt(i, SYMBOL_COL_IDX); 
            if (stockSymbol.equals(rowSymbol)) {
                stockTableModel.setValueAt(new Integer(todayPrice), i, CURRENT_PRICE_COL_IDX);
                stockTableModel.setValueAt(priceChangePct, i, PRICE_CHANGE_PCT_COL_IDX);
                return;
            }
        }
    }
    
    /** does updates needed when game ends */
    private void doEndGameTasks() {
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    doEndGameTasks();
                }
            });
        }
        stockTableModel.setRowCount(0);
        buyButton.setEnabled(false);
    }

    private class BuyStockActionListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            if (!game.isActive()) {
                return;
            }

            String symbol = (String) stockTable.getValueAt(stockTable.getSelectedRow(), SYMBOL_COL_IDX);
            int currentPrice = game.getDataProvider().getPrice(symbol,game.getCurrentDay());
            String formattedPrice = priceFormatter.format(currentPrice);
            
            // NotifyDescriptor.InputLine is similar to how you might prompt the user with JOptionPane 
            String title = NbBundle.getMessage(ExchangeListingComponent.class, "LBL_BuyStockDialogTitle");
            String msg = NbBundle.getMessage(ExchangeListingComponent.class,
                    "MSG_HowManySharesToBuy", symbol, formattedPrice);            
            NotifyDescriptor.InputLine inputDialog = new NotifyDescriptor.InputLine(msg, title);            
            if (DialogDisplayer.getDefault().notify(inputDialog) == NotifyDescriptor.OK_OPTION) {
                String qtyString = inputDialog.getInputText();
                if (qtyString != null) {                
                    try {
                        int numShares = Integer.parseInt(qtyString);
                        attemptPurchase(symbol, numShares);
                    } catch (NumberFormatException nfe) {
                        // show a dialog box with the error message
                        String errMsg = NbBundle.getMessage(ExchangeListingComponent.class,
                                "MSG_InvalidQty");
                        NotifyDescriptor descriptor = new NotifyDescriptor.Message(errMsg,
                                NotifyDescriptor.ERROR_MESSAGE);
                        DialogDisplayer.getDefault().notify(descriptor);
                    }
                }
            } else {
                // gently inform them in case cancellation was accidental
                String cancelMsg = NbBundle.getMessage(
                        ExchangeListingComponent.class, "MSG_BuyTxCancelled");
                StatusDisplayer.getDefault().setStatusText(cancelMsg);
            }
        }
    }

    // keeps track of price changes, so we can update the display
    private class PriceListener implements StockPriceListener {
        public void stockPriceUpdated(StockPriceChangeEvent event) {
            updateStockListing(event.getSymbol());
        }
    }
    
    // keeps track of game start/stop, so we can reinitialize
    private class CurrentGameListener implements GameListener {

        public void gameStarted(GameChangeEvent event) {
            game.getDataProvider().addStockPriceListener(priceListener);            
            populateListingTable();
        }

        public void gameEnded(GameChangeEvent event) {
            doEndGameTasks();
        }
        
        public void dayChanged(GameChangeEvent event) { }
    }
}
