package com.tomwheeler.stocktrader.ui.gamectrl;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.NumberFormat;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.*;

import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import org.openide.util.NbBundle;

import com.tomwheeler.stocktrader.api.*;
import com.tomwheeler.stocktrader.api.events.*;

/**
 * <p>
 * This class provides a component for controlling the stocktrader game. It
 * provides the main part of the UI rendered by the GameControllerTopComponent.
 * Aside from how user notification resource bundles are handled, there's 
 * nothing specific to NetBeans in this class.
 * </p>
 * 
 * @author Tom Wheeler
 */
public class GameControllerComponent extends JPanel {

    private static NumberFormat currencyFormatter;

    private JLabel totalLabel, cashLabel, stockLabel, gameDayLabel;
    private JButton toggleGameButton, nextDayButton;
    private StockPriceListener priceListener;
    private StockAccountListener accountListener;
    private Game game;    

    public GameControllerComponent(final Game game) {
        super(new BorderLayout());        
        currencyFormatter = NumberFormat.getCurrencyInstance();        
        this.game = game;

        String startLabel = NbBundle.getMessage(GameControllerComponent.class, "LBL_StartGameButton");
        toggleGameButton = new JButton(startLabel);
        toggleGameButton.addActionListener(new ToggleGameListener());

        String nextDayLabel = NbBundle.getMessage(GameControllerComponent.class, "LBL_NextDayButton");
        nextDayButton = new JButton(nextDayLabel);
        nextDayButton.setEnabled(false);
        nextDayButton.addActionListener(new NextDayListener());
        
        JPanel controlPanel = new JPanel(new BorderLayout());
        controlPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
        controlPanel.add(toggleGameButton, BorderLayout.WEST);
        controlPanel.add(nextDayButton, BorderLayout.EAST);        
        add(controlPanel, BorderLayout.NORTH);
        
        JPanel acctInfoPanel = new JPanel(new GridLayout(4,1));
        String borderText = NbBundle.getMessage(GameControllerComponent.class, "LBL_AccountInfoBorder");
        acctInfoPanel.setBorder(BorderFactory.createTitledBorder(borderText));
        
        gameDayLabel = new JLabel();
        cashLabel = new JLabel();
        totalLabel = new JLabel();
        stockLabel = new JLabel();
        
        acctInfoPanel.add(gameDayLabel);
        acctInfoPanel.add(cashLabel);        
        acctInfoPanel.add(stockLabel);
        acctInfoPanel.add(totalLabel);
        
        updateDisplayedValues();
        add(acctInfoPanel, BorderLayout.CENTER);
    }
    
    private void startGame() {
        String endLabel = NbBundle.getMessage(GameControllerComponent.class, "LBL_EndGameButton");
        toggleGameButton.setText(endLabel);
        game.startGame();
        game.getDataProvider().addStockPriceListener(new PriceListener());
        game.getAccount().addStockAccountListener(new AccountListener());
        
        nextDayButton.setEnabled(true);
        updateDisplayedValues();
    }
    
    private void endGame() {
        String startLabel = NbBundle.getMessage(GameControllerComponent.class, "LBL_StartGameButton");
        toggleGameButton.setText(startLabel);
        nextDayButton.setEnabled(false);
        
        int score = game.endGame();
        updateDisplayedValues();

        // display the user's score in a dialog
        String msg = NbBundle.getMessage(GameControllerComponent.class, "MSG_ShowEndScore",
                new Integer(score));
        NotifyDescriptor descriptor =
            new NotifyDescriptor.Message(msg, NotifyDescriptor.INFORMATION_MESSAGE);
        DialogDisplayer.getDefault().notify(descriptor);
    }
    
    private void nextDay() {
        game.nextDay();
        updateDisplayedValues();
    }
    
    /**
     * <p>
     * Convenience method to calculate and return the current value of all stocks
     * in the account. 
     * </p>
     * 
     * @return the current value of all stocks in the account
     */
    private int getCurrentPortfolioValue() {
        int portfolioValue = 0;
        
        int currentDay = game.getCurrentDay();
        StockDataProvider provider = game.getDataProvider();
        StockAccount acct = game.getAccount();
        
        Collection stocks = acct.getStocks();
        for (Iterator iter = stocks.iterator(); iter.hasNext();) {
            Stock stock = (Stock) iter.next();
            int qty = acct.getQuantity(stock);
            if (qty > 0) {
                int currentPrice = provider.getPrice(stock.getSymbol(), currentDay);               
                portfolioValue += currentPrice * qty;                
            }
        }
        return portfolioValue;
    }

    private void updateDisplayedValues() {
        if (SwingUtilities.isEventDispatchThread()) {
            Class c = GameControllerComponent.class;
            if (game.isActive()) {
                String msg = NbBundle.getMessage(c, "LBL_CurrentDay", 
                        new Integer(game.getCurrentDay()));
                gameDayLabel.setText(msg);
                
                StockAccount acct = game.getAccount();
                int stock = getCurrentPortfolioValue();                
                stockLabel.setText(NbBundle.getMessage(c, "LBL_StockValueStarted",
                        currencyFormatter.format(stock)));
                
                int cash = game.getAccount().getCashBalance();
                cashLabel.setText(NbBundle.getMessage(c, "LBL_CashValueStarted",
                        currencyFormatter.format(cash)));

                int total = stock + cash;                
                totalLabel.setText(NbBundle.getMessage(c, "LBL_TotalValueStarted",
                        currencyFormatter.format(total)));
            } else {
                gameDayLabel.setText(NbBundle.getMessage(c,"LBL_NoGameStarted"));                
                totalLabel.setText(NbBundle.getMessage(c, "LBL_TotalValueNotStarted"));
                cashLabel.setText(NbBundle.getMessage(c, "LBL_CashValueNotStarted"));
                stockLabel.setText(NbBundle.getMessage(c, "LBL_StockValueNotStarted"));
            }
        } else {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    updateDisplayedValues();
                }
            });
        }
    }

    private class ToggleGameListener implements ActionListener {
        public void actionPerformed(ActionEvent ae) {
            if (game.isActive()) {
                endGame();
            } else {
                startGame(); 
            }
        }
    }
    
    private class NextDayListener implements ActionListener {
        public void actionPerformed(ActionEvent ae) {
            if (game.isActive()) {
                nextDay();
            }
        }
    }
    
    private class AccountListener implements StockAccountListener {
        public void stockPurchased(StockAccountChangeEvent event) {
            updateDisplayedValues();            
        }

        public void stocksSold(StockAccountChangeEvent event) {
            updateDisplayedValues();            
        }
    }
    
    // so we can update when prices change
    private class PriceListener implements StockPriceListener {
        public void stockPriceUpdated(StockPriceChangeEvent event) {
            // impl detail: this is not efficient, since we recalculate the stock 
            // and cash values with each change.  In reality, they change in batch
            // at the start of each day, but the GUI should not know this detail.
            updateDisplayedValues();
        }
    }
}
