/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2005 Sun
 * Microsystems, Inc. All Rights Reserved.
 */
package com.tomwheeler.stocktrader.ui.portfolio;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.NumberFormat;
import java.util.Collection;
import java.util.Vector;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableColumnModel;

import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import org.openide.awt.StatusDisplayer;
import org.openide.util.NbBundle;

import com.tomwheeler.stocktrader.api.*;
import com.tomwheeler.stocktrader.api.events.*;
import com.tomwheeler.stocktrader.ui.StockListTableModel;
import com.tomwheeler.stocktrader.ui.renderer.CurrencyCellRenderer;

/**
 * <p>
 * This class provides a component that can display stock account data.  
 * It provides the main part of the UI rendered by the PortfolioTopComponent.  
 * Aside from how user notification and resource bundles are handled, there's 
 * nothing specific to NetBeans in this class.
 * </p> 
 * 
 * @author Tom Wheeler
 */
public class PortfolioComponent extends JPanel {
    
    private static NumberFormat priceFormatter = NumberFormat.getCurrencyInstance();

    private static final int SYMBOL_COL_IDX = 0;
    private static final int CURRENT_PRICE_COL_IDX = 1;
    private static final int QTY_COL_IDX = 2;
    private static final int TOTAL_VALUE_COL_IDX = 3;

    private static NumberFormat currencyFormatter;
    
    private JTable stockTable;
    private StockListTableModel stockTableModel;
    private JButton sellButton;
    
    private Game game;    
    private StockPriceListener priceListener;
    private StockAccountListener stockAccountListener;
    private GameListener gameListener;

    public PortfolioComponent(final Game game) {
        super(new BorderLayout());
        
        currencyFormatter = NumberFormat.getCurrencyInstance();
        
        this.game = game;
        
        stockTableModel = new StockListTableModel();
        stockTableModel.addColumn(NbBundle.getMessage(PortfolioComponent.class,
                "LBL_SymbolCol"));
        stockTableModel.addColumn(NbBundle.getMessage(PortfolioComponent.class,
                "LBL_PriceCol"));
        stockTableModel.addColumn(NbBundle.getMessage(PortfolioComponent.class,
                "LBL_QtyCol"));
        stockTableModel.addColumn(NbBundle.getMessage(PortfolioComponent.class,
                "LBL_TotalCol"));

        stockTable = new JTable(stockTableModel);
        ListSelectionModel singleSelectionModel = new DefaultListSelectionModel();
        singleSelectionModel.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        stockTable.setSelectionModel(singleSelectionModel);
        stockTable.getTableHeader().setReorderingAllowed(false); 
        
        TableColumnModel colModel = stockTable.getColumnModel();
        colModel.getColumn(TOTAL_VALUE_COL_IDX).setCellRenderer(new CurrencyCellRenderer());

        JScrollPane scroller = new JScrollPane(stockTable,
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

        sellButton = new JButton(NbBundle.getMessage(PortfolioComponent.class,"LBL_SellStockButton"));
        sellButton.setEnabled(false);  // enabled upon selection
        
        stockTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
            if (game.isActive()) {
                    boolean isEnabled = stockTable.getSelectedRow() != -1;
                    sellButton.setEnabled(isEnabled);                    
                }
            }            
        });
        
        sellButton.addActionListener(new SellStockActionListener());
        
        game.addGameListener(new CurrentGameListener());
        priceListener = new PriceListener();
        stockAccountListener = new AccountListener();

        add(sellButton, BorderLayout.SOUTH);
        add(scroller, BorderLayout.CENTER);        
    }

    /** attempts to sell the specified stock and handles confirmation/error message display */
    private void attemptSale(String symbol, int qty) {
        try {
            String confMsg = game.getAccount().sell(symbol, qty);
            StatusDisplayer.getDefault().setStatusText(confMsg);
        } catch (IllegalArgumentException iae) {
            String msg = iae.getLocalizedMessage();
            NotifyDescriptor descriptor = new NotifyDescriptor.Message(msg,
                    NotifyDescriptor.ERROR_MESSAGE);
            DialogDisplayer.getDefault().notify(descriptor);
        }
    }

    /** does updates needed when game ends */
    private void doEndGameTasks() {
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    doEndGameTasks();
                }
            });
        }
        stockTableModel.setRowCount(0);
        sellButton.setEnabled(false);
    }
    
    /** updates (or adds) data associated with the specified stock in the table */
    private void dataUpdated(String stockSymbol) {
        if (! game.isActive()) {
            return;
        }

        StockAccount acct = game.getAccount();
        StockExchange exchange = game.getExchange();
        StockDataProvider provider = game.getDataProvider();
        
        Stock stock = game.getExchange().getStock(stockSymbol);
        
        int qty = acct.getQuantity(stock);
        int todayPrice = provider.getPrice(stockSymbol, game.getCurrentDay());        
        int totalValue = todayPrice * qty;
        Collection allStocks = acct.getStocks();
        int numRows = stockTableModel.getRowCount();
        for (int i = 0; i < numRows; i++) {
            String rowSymbol = (String)stockTableModel.getValueAt(i, SYMBOL_COL_IDX); 
            if (stockSymbol.equals(rowSymbol)) {
                if (qty <= 0) {
                    // we did own it, but probably just sold it all
                    stockTableModel.removeRow(i);
                } else {
                    stockTableModel.setValueAt(new Integer(todayPrice), i, CURRENT_PRICE_COL_IDX);
                    stockTableModel.setValueAt(new Integer(qty), i, QTY_COL_IDX);
                    stockTableModel.setValueAt(new Integer(totalValue), i, TOTAL_VALUE_COL_IDX);                    
                }
                return;
            }
        }
        
        if (qty <= 0) {
            // we don't own it
            return; 
        }
        
        // if we reach this point, then we don't currently hold the stock and must add it
        Vector rowData = new Vector();
        rowData.insertElementAt(stockSymbol, SYMBOL_COL_IDX);
        rowData.insertElementAt(new Integer(todayPrice), CURRENT_PRICE_COL_IDX);
        rowData.insertElementAt(new Integer(qty), QTY_COL_IDX);
        rowData.insertElementAt(new Integer(totalValue), TOTAL_VALUE_COL_IDX);
        stockTableModel.addRow(rowData);
    }
    
    private class SellStockActionListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            if (!game.isActive()) {
                return;
            }

            String symbol = (String) stockTable.getValueAt(stockTable.getSelectedRow(), SYMBOL_COL_IDX);
            int currentPrice = game.getDataProvider().getPrice(symbol,game.getCurrentDay());
            String formattedPrice = priceFormatter.format(currentPrice);
            
            // NotifyDescriptor.InputLine is similar to how you might prompt the user with JOptionPane 
            String title = NbBundle.getMessage(PortfolioComponent.class, "LBL_SellStockDialogTitle");
            String msg = NbBundle.getMessage(PortfolioComponent.class,
                    "MSG_HowManySharesToSell", symbol, formattedPrice);
            NotifyDescriptor.InputLine inputDialog = new NotifyDescriptor.InputLine(msg, title);

            // show the dialog and check which button they pushed
            if (DialogDisplayer.getDefault().notify(inputDialog) == NotifyDescriptor.OK_OPTION) {
                String qtyString = inputDialog.getInputText();
                if (qtyString != null) {                
                    try {
                        int numShares = Integer.parseInt(qtyString);
                        attemptSale(symbol, numShares);
                    } catch (NumberFormatException nfe) {
                        // show a dialog box with the error message
                        String errMsg = NbBundle.getMessage(PortfolioComponent.class, "MSG_InvalidQty");
                        NotifyDescriptor descriptor = new NotifyDescriptor.Message(errMsg,
                                NotifyDescriptor.ERROR_MESSAGE);
                        DialogDisplayer.getDefault().notify(descriptor);
                    }
                }
            } else {
                // gently inform them in case cancellation was accidental
                String cancelMsg = NbBundle.getMessage(PortfolioComponent.class, "MSG_SellTxCancelled");
                StatusDisplayer.getDefault().setStatusText(cancelMsg);                
            }
        }
    }

    // so we can update when stock is bought or sold
    private class AccountListener implements StockAccountListener {
        public void accountChanged(StockAccountChangeEvent event) {
            dataUpdated(event.getDescription());
        }
        
        public void stockPurchased(StockAccountChangeEvent event) {
            dataUpdated(event.getSymbol());
        }

        public void stocksSold(StockAccountChangeEvent event) {
            dataUpdated(event.getSymbol());            
        }
    }
    
    // so we can update when prices change
    private class PriceListener implements StockPriceListener {
        public void stockPriceUpdated(StockPriceChangeEvent event) {
            dataUpdated(event.getSymbol());
        }
    }
    
    // keeps track of game start/stop, so we can reinitialize
    private class CurrentGameListener implements GameListener {

        public void gameStarted(GameChangeEvent event) {
            game.getDataProvider().addStockPriceListener(priceListener);
            game.getAccount().addStockAccountListener(stockAccountListener);    
        }

        public void gameEnded(GameChangeEvent event) {
            doEndGameTasks();
        }
        
        public void dayChanged(GameChangeEvent event) {}
    }
}
